/*

Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "GraphicsContext.h"
#include "PngUtil.h"

#include "TextureCache.h"
#include "Resources\resource.h"
#include "D3DRender.h"

#include "Debug\DBGConsole.h"

#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)  { if(p) { (p)->Release();     (p)=NULL; } }
#endif

LPDIRECT3DDEVICE8 g_pD3DDev = NULL;


//-----------------------------------------------------------------------------
// Error codes
//-----------------------------------------------------------------------------
enum APPMSGTYPE { MSG_NONE, MSGERR_APPMUSTEXIT, MSGWARN_SWITCHEDTOREF };

#define D3DAPPERR_NODIRECT3D          0x82000001
#define D3DAPPERR_NOCOMPATIBLEDEVICES 0x82000003
#define D3DAPPERR_NOWINDOWABLEDEVICES 0x82000004
#define D3DAPPERR_NOHALTHISMODE       0x82000009
#define D3DAPPERR_NONZEROREFCOUNT     0x8200000a
#define D3DAPPERR_RESIZEFAILED        0x8200000c


//*****************************************************************************
// 
//*****************************************************************************
class CMode
{
public:
	DWORD      Width;      // Screen width in this mode
	DWORD      Height;     // Screen height in this mode
	D3DFORMAT  Format;     // Pixel format in this mode
	DWORD      dwBehavior; // Hardware / Software / Mixed vertex processing
	D3DFORMAT  DepthStencilFormat; // Which depth/stencil format to use with this mode
};

//*****************************************************************************
// 
//*****************************************************************************
class CDevice
{
public:
	void	Initialise( LPDIRECT3D8 pd3d, u32 adapter, D3DDEVTYPE dev_type )
	{
		DeviceType     = dev_type;
		pd3d->GetDeviceCaps( adapter, dev_type, &d3dCaps );
		strDesc        = dev_type == D3DDEVTYPE_HAL ? "HAL" : "REF";
		dwNumModes     = 0;
		dwCurrentMode  = 0;
		bCanDoWindowed = false;
		bWindowed      = false;
		MultiSampleType = D3DMULTISAMPLE_NONE;
	}
	
	D3DDEVTYPE		GetType() const					{ return DeviceType; }
	D3DCAPS8		GetCaps() const					{ return d3dCaps; }
	const TCHAR *	GetDescription( ) const			{ return strDesc; }
	
	D3DMULTISAMPLE_TYPE GetMultiSampleType() const	{ return MultiSampleType; }
	void			SetMultiSampleType( D3DMULTISAMPLE_TYPE type )	{ MultiSampleType = type; }
	
	bool			CanDoWindowed() const			{ return bCanDoWindowed; }
	void			SetCanDoWindowed( bool can_do )	{ bCanDoWindowed = can_do; }
	
	bool			GetWindowed() const				{ return bWindowed; }
	void			SetWindowed( bool windowed )	{ bWindowed = windowed; }
	
	CMode *			GetMode( u32 i )				{ return &modes[ i ]; }
	const CMode *	GetMode( u32 i ) const			{ return &modes[ i ]; }
	
	CMode *			GetCurrentMode()				{ return &modes[ dwCurrentMode ]; }
	const CMode *	GetCurrentMode() const			{ return &modes[ dwCurrentMode ]; }
	
	u32				GetNumModes() const				{ return dwNumModes; }
	
	u32				GetCurrentModeIdx() const		{ return dwCurrentMode; }
	void			SetCurrentMode( u32 i )			{ dwCurrentMode = i; }
	
	void			AddMode( D3DDISPLAYMODE mode, DWORD behaviour, D3DFORMAT depth_fmt )
	{
		modes[ dwNumModes ].Width      = mode.Width;
		modes[ dwNumModes ].Height     = mode.Height;
		modes[ dwNumModes ].Format     = mode.Format;
		modes[ dwNumModes ].dwBehavior = behaviour;
		modes[ dwNumModes ].DepthStencilFormat = depth_fmt;
		dwNumModes++;
	}
	
	void	SelectDefaultMode()
	{
		// Select any 640x480 mode for default (but prefer a 16-bit mode)
		for( u32 i = 0; i < GetNumModes(); i++ )
		{
			const CMode * pMode = GetMode( i );
			
			if( pMode->Width==640 && pMode->Height==480 )
			{
				SetCurrentMode( i );
				
				if( pMode->Format == D3DFMT_R5G6B5 ||
					pMode->Format == D3DFMT_X1R5G5B5 ||
					pMode->Format == D3DFMT_A1R5G5B5 )
				{
					return;
				}
			}
		}
	}
	
	
	
private:
	// Device data
	D3DDEVTYPE   DeviceType;      // Reference, HAL, etc.
	D3DCAPS8     d3dCaps;         // Capabilities of this device
	const TCHAR* strDesc;         // Name of this device
	bool         bCanDoWindowed;  // Whether this device can work in windowed mode
	
	// Modes for this device
	DWORD        dwNumModes;
	CMode		 modes[150];
	
	// Current state
	DWORD        dwCurrentMode;
	bool         bWindowed;
	D3DMULTISAMPLE_TYPE MultiSampleType;
};


//*****************************************************************************
// 
//*****************************************************************************
class CAdapter
{
	public:

		CAdapter() : mpD3D( NULL )
		{

		}

		virtual ~CAdapter()
		{ 
			// Release
		}

		typedef std::vector< D3DDISPLAYMODE > ModeVector;
		typedef std::vector< D3DFORMAT >	  FormatVector;
		
		void	Initialise( LPDIRECT3D8 pd3d, u32 adapter )
		{
			mpD3D = pd3d;
			mAdapterIndex = adapter;
			
			mpD3D->GetAdapterIdentifier( mAdapterIndex, 0, &d3dAdapterIdentifier );
			mpD3D->GetAdapterDisplayMode( mAdapterIndex, &d3ddmDesktop );
			
			dwNumDevices = 0;
			dwCurrentDevice = 0;

			//
			// Enumerate all display modes on this adapter
			//
			ModeVector		modes;
			FormatVector    formats;
			
			BuildModesAndFormatsList( modes, formats );
			
			// Sort the list of display modes (by format, then width, then height)
			qsort( &modes[0], modes.size(), sizeof(D3DDISPLAYMODE), SortModesCallback );
			
			AddModesForDevice( D3DDEVTYPE_HAL, modes, formats );
			AddModesForDevice( D3DDEVTYPE_REF, modes, formats );
			AddModesForDevice( D3DDEVTYPE_SW, modes, formats );
		}

		// Check if the mode already exists (to filter out refresh rates)
		bool	ModeExists( const ModeVector & modes, const D3DDISPLAYMODE & mode ) const
		{
			for( u32 i = 0; i < modes.size(); i++ )
			{
				if( ( modes[i].Width  == mode.Width  ) &&
					( modes[i].Height == mode.Height ) &&
					( modes[i].Format == mode.Format ) )
				{
					return true;
				}
			}
			return false;
		}

		// Check if the mode's format already exists
		bool	FormatExists( const FormatVector & formats, D3DFORMAT format ) const
		{
			for( u32 i = 0; i < formats.size(); i++ )
			{
				if( formats[i] == format )
				{
					return true;
				}
			}
			return false;
		}
		
		void	BuildModesAndFormatsList( ModeVector & modes, FormatVector & formats ) const
		{
			D3DDISPLAYMODE display_mode;
			u32			num_adapter_modes = mpD3D->GetAdapterModeCount( GetAdapterIndex() );
			
			// Add the adapter's current desktop format to the list of formats
			formats.push_back( GetDesktopDisplayMode().Format );
			
			for( u32 i = 0; i < num_adapter_modes; i++ )
			{
				// Get the display mode attributes
				mpD3D->EnumAdapterModes( GetAdapterIndex(), i, &display_mode );

				// Ignore the refresh rate
				display_mode.RefreshRate = 0;
				
				// Filter out low-resolution modes
				if( display_mode.Width  < 640 || display_mode.Height < 480 )
					continue;
				
				// Check if the mode already exists (to filter out refresh rates)
				if( !ModeExists( modes, display_mode ) )
				{
					modes.push_back( display_mode );
					
					// If the format is new, add it to the list
					if( !FormatExists( formats, display_mode.Format ) )
					{
						formats.push_back( display_mode.Format );
					}
				}
			}
		}

		bool	AddModesForDevice( D3DDEVTYPE device_type, const ModeVector & modes, const FormatVector & formats )
		{
			// Fill in device info
			CDevice			device;
			
			device.Initialise( mpD3D, GetAdapterIndex(), device_type );
			
			// Examine each format supported by the adapter to see if it will
			// work with this device and meets the needs of the application.
			BOOL  bFormatConfirmed[20];
			DWORD dwBehavior[20];
			D3DFORMAT fmtDepthStencil[20];
			
			for( DWORD f=0; f<formats.size(); f++ )
			{
				bFormatConfirmed[f] = FALSE;
				fmtDepthStencil[f] = D3DFMT_UNKNOWN;
				
				// Skip formats that cannot be used as render targets on this device
				if( FAILED( mpD3D->CheckDeviceType( GetAdapterIndex(), device.GetType(), formats[f], formats[f], FALSE ) ) )
					continue;
								
				// Confirm the device/format for HW vertex processing
				if( device.GetCaps().DevCaps&D3DDEVCAPS_HWTRANSFORMANDLIGHT )
				{
					if( device.GetCaps().DevCaps&D3DDEVCAPS_PUREDEVICE )
					{
						dwBehavior[f] = D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE;	
						bFormatConfirmed[f] = TRUE;
					}
					
					if ( !bFormatConfirmed[f] )
					{
						dwBehavior[f] = D3DCREATE_HARDWARE_VERTEXPROCESSING;
						bFormatConfirmed[f] = TRUE;
					}
				}
				
				// Confirm the device/format for SW vertex processing
				if( !bFormatConfirmed[f] )
				{
					dwBehavior[f] = D3DCREATE_SOFTWARE_VERTEXPROCESSING;	
					bFormatConfirmed[f] = TRUE;
				}
				
				
				// Find a suitable depth/stencil buffer format for this device/format
				if( bFormatConfirmed[f] )
				{
					if( !FindDepthStencilFormat( device.GetType(), formats[f], &fmtDepthStencil[f] ) )
					{
						bFormatConfirmed[f] = FALSE;
					}
				}
			}
			
			// Add all enumerated display modes with confirmed formats to the
			// device's list of valid modes
			for( DWORD m=0L; m<modes.size(); m++ )
			{
				for( DWORD f=0; f<formats.size(); f++ )
				{
					if( modes[m].Format == formats[f] )
					{
						if( bFormatConfirmed[f] )
						{
							// Add this mode to the device's list of valid modes
							device.AddMode( modes[m], dwBehavior[f], fmtDepthStencil[f] );
						}
					}
				}
			}
			
			// If valid modes were found, keep this device
			if( device.GetNumModes() > 0 )
			{
				//
				// Select the default mode for this device
				//
				device.SelectDefaultMode();
				
				// Check if the device is compatible with the desktop display mode
				// (which was added initially as formats[0])
				if( bFormatConfirmed[0] && (device.GetCaps().Caps2 & D3DCAPS2_CANRENDERWINDOWED) )
				{
					device.SetCanDoWindowed( true );
					device.SetWindowed( true );
				}
			
				AddDevice( device );
				return true;
			}
			else
			{
				return false;
			}
		}


		
		u32								GetAdapterIndex() const			{ return mAdapterIndex; }
		
		const D3DADAPTER_IDENTIFIER8 &	GetAdapterID( ) const			{ return d3dAdapterIdentifier; }
		const D3DDISPLAYMODE &			GetDesktopDisplayMode( ) const	{ return d3ddmDesktop; }
		
		void							SetAdapterID( const D3DADAPTER_IDENTIFIER8 & adapter_id )		{ d3dAdapterIdentifier = adapter_id; }
		void							SetDesktopDisplayMode( const D3DDISPLAYMODE & display_mode )	{ d3ddmDesktop = display_mode; }
		
		CDevice *						GetDevice( u32 i )				{ return &devices[ i ]; }
		const CDevice *					GetDevice( u32 i ) const		{ return &devices[ i ]; }
		
		u32								GetCurrentDeviceIdx( )			{ return dwCurrentDevice; }
		CDevice *						GetCurrentDevice( )				{ return &devices[ dwCurrentDevice ]; }
		const CDevice *					GetCurrentDevice( ) const		{ return &devices[ dwCurrentDevice ]; }
		
		void							SetCurrentDevice( u32 i )		{ dwCurrentDevice = i; }
		
		u32								GetNumDevices( ) const			{ return dwNumDevices; }
		
		void							AddDevice( const CDevice & device )
		{
			devices[ dwNumDevices ] = device;
			dwNumDevices++;
		}


	private:
		static int _cdecl SortModesCallback( const VOID* arg1, const VOID* arg2 )
		{
			D3DDISPLAYMODE* p1 = (D3DDISPLAYMODE*)arg1;
			D3DDISPLAYMODE* p2 = (D3DDISPLAYMODE*)arg2;
			
			if( p1->Format > p2->Format )   return -1;
			if( p1->Format < p2->Format )   return +1;
			if( p1->Width  < p2->Width )    return -1;
			if( p1->Width  > p2->Width )    return +1;
			if( p1->Height < p2->Height )   return -1;
			if( p1->Height > p2->Height )   return +1;
			
			return 0;
		}
		
		bool FindDepthStencilFormat( D3DDEVTYPE device_type,
									 D3DFORMAT target_format,
									 D3DFORMAT* p_depth_fmt )
		{
			#ifndef ARRAYSIZE
			#define ARRAYSIZE(x)	(  sizeof( (x)  ) / sizeof(  (x)[0]  )   )
			#endif

			static const D3DFORMAT	format_list[] = { D3DFMT_D16, D3DFMT_D24X8, D3DFMT_D24S8, D3DFMT_D24X4S4, D3DFMT_D32 };

			for ( u32 i = 0; i < ARRAYSIZE( format_list ); i++ )
			{
				D3DFORMAT format = format_list[ i ];

				if( SUCCEEDED( mpD3D->CheckDeviceFormat( GetAdapterIndex(), device_type, target_format, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, format ) ) )
				{
					if( SUCCEEDED( mpD3D->CheckDepthStencilMatch( GetAdapterIndex(), device_type, target_format, target_format, format ) ) )
					{
						*p_depth_fmt = format;
						return true;
					}
				}
			}

			return false;
		}


	private:
		LPDIRECT3D8		mpD3D;
		u32				mAdapterIndex;
		// Adapter data
		D3DADAPTER_IDENTIFIER8 d3dAdapterIdentifier;
		D3DDISPLAYMODE d3ddmDesktop;      // Desktop display mode for this adapter
		
		// Devices for this adapter
		DWORD          dwNumDevices;
		CDevice		   devices[5];
		
		// Current state
		DWORD          dwCurrentDevice;
};



// Implementation
class IGraphicsContext : public CGraphicsContext
{
public:
	virtual ~IGraphicsContext();
	
	BOOL Ready() { return m_bReady; }
	
	void SetReady(BOOL bReady) { m_bReady = bReady; }
	
	D3DCAPS8 GetCaps() const { return m_d3dCaps; }
	
	void Cleanup3DEnvironment();
	
	LPDIRECT3DDEVICE8 GetD3DDev();
	void Clear(DWORD dwFlags);
	HRESULT UpdateFrame();
	HRESULT GetBufferSize(DWORD * pdwWidth, DWORD * pdwHeight);
	HRESULT Initialise(HWND hWnd, HWND hWndStatus, DWORD dwWidth, DWORD dwHeight, BOOL bWindowed );
	
	LRESULT OnSize( WPARAM wParam );
	LRESULT OnExitSizeMove();
	
	void OnChangeDevice();
	void ToggleFullscreen();
	
	void DumpScreenShot();
	
	
protected:
	friend class CUnique< CGraphicsContext >;
	
	IGraphicsContext();
	
	HRESULT BuildDeviceList();
	
	void BuildAdapterList();
	
	
	HRESULT DisplayErrorMsg( HRESULT hr, DWORD dwType );
	
	HRESULT Create( BOOL bWindowed );
	HRESULT Initialize3DEnvironment();
	HRESULT Resize3DEnvironment();
	HRESULT DoToggleFullscreen();
	HRESULT ForceWindowed();
	HRESULT AdjustWindowForChange();
	
	HRESULT UserSelectNewDevice();
	
	static INT_PTR CALLBACK SelectDeviceProc( HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam );
	
	
	LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
	
	VOID Pause( BOOL bPause );
	
	HRESULT InitDeviceObjects();
	HRESULT RestoreDeviceObjects();
	HRESULT InvalidateDeviceObjects();
	HRESULT DeleteDeviceObjects();
	
	
	
	HWND				m_hWnd;
	HWND				m_hWndStatus;
	HWND				m_hWndFocus;
	BOOL				m_bReady;
	BOOL				m_bActive;
	
	LPDIRECT3D8 		m_pD3D;
	LPDIRECT3DDEVICE8	m_pd3dDevice;
	D3DCAPS8			m_d3dCaps;			 // Caps for the device
	D3DSURFACE_DESC 	m_d3dsdBackBuffer;	 // Surface desc of the backbuffer
	DWORD				m_dwWindowStyle;	 // Saved window style for mode switches
	DWORD				m_dwWindowExStyle;	 // Saved window style for mode switches
	DWORD				m_dwStatusWindowStyle;	   // Saved window style for mode switches
	
	D3DPRESENT_PARAMETERS m_d3dpp;
	
	CAdapter			m_Adapters[10];
	DWORD				m_dwNumAdapters;
	
	DWORD				m_dwAdapter;
	
	BOOL				m_bWindowed;
	
	RECT				m_rcWindowClient;
	RECT				m_rcWindowBounds;

	CHAR				m_strDeviceStats[90];// String to hold D3D device stats
};



HRESULT CUnique< CGraphicsContext >::Create()
{	
	IGraphicsContext * pNewInstance;
	
	_ASSERTE(m_pInstance == NULL);
	
	pNewInstance = new IGraphicsContext();
	if (!pNewInstance)
	{
		return E_OUTOFMEMORY;
	}
	
	m_pInstance = pNewInstance;
	
	return S_OK;
};


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IGraphicsContext::IGraphicsContext() :
	m_bReady(FALSE), 
	m_bActive(FALSE),
	m_pd3dDevice(NULL),
	m_pD3D(NULL),
	m_dwNumAdapters(0),
	m_dwAdapter(0),			// Default Adapter
	m_bWindowed(TRUE),
	m_hWnd(NULL),			// TODO: Assign window handle somewhere
	m_hWndFocus(NULL)		// TODO: Assign window handle somewhere
{
	m_strDeviceStats[0] = '\0';
}

//*****************************************************************************
//
//*****************************************************************************
IGraphicsContext::~IGraphicsContext()
{
	CGraphicsContext::Get()->Cleanup3DEnvironment();
}


//*****************************************************************************
//
//*****************************************************************************
LPDIRECT3DDEVICE8 IGraphicsContext::GetD3DDev()
{
	// Assert we're locked
	return m_pd3dDevice;
}

//*****************************************************************************
//
//*****************************************************************************
void IGraphicsContext::Clear(DWORD dwFlags)
{
	Lock();
	if (m_pd3dDevice != NULL)
		m_pd3dDevice->Clear(0, NULL, dwFlags, 0xff000000, 1.0f, 0);
	Unlock();
}

//*****************************************************************************
//
//*****************************************************************************
HRESULT IGraphicsContext::UpdateFrame()
{
	HRESULT hr;
	
	//	if ( !m_bReady )
	//		return E_FAIL;
	
	
	Lock();
	if (m_pd3dDevice == NULL)
	{
		hr = E_FAIL;
	}
	else
	{
		// Test the cooperative level to see if it's okay to render
		if( FAILED( hr = m_pd3dDevice->TestCooperativeLevel() ) )
		{
			// If the device was lost, do not render until we get it back
			if( hr == D3DERR_DEVICELOST )
			{
				hr = S_OK;
				goto exit;
			}
			
			// Check if the device needs to be resized.
			if( hr == D3DERR_DEVICENOTRESET )
			{
				// If we are windowed, read the desktop mode and use the same format for
				// the back buffer
				if( m_bWindowed )
				{
					D3DDISPLAYMODE display_mode;
					m_pD3D->GetAdapterDisplayMode( m_dwAdapter, &display_mode );
					
					m_Adapters[m_dwAdapter].SetDesktopDisplayMode( display_mode );
					
					m_d3dpp.BackBufferFormat = display_mode.Format;
				}
				
				if( FAILED( hr = Resize3DEnvironment() ) )
				{
					goto exit;
				}
			}
			
			// return hr
			goto exit;
		}
		
		if ( IsWindow( m_hWndStatus ) && m_bWindowed )
		{		
			hr = m_pd3dDevice->Present( &m_rcWindowClient, &m_rcWindowClient, NULL, NULL );
		}
		else
		{
			hr = m_pd3dDevice->Present( NULL, NULL, NULL, NULL );
		}
	}
exit:
	
	Unlock();
	
	return hr;
}


//*****************************************************************************
//
//*****************************************************************************
void IGraphicsContext::DumpScreenShot()
{
	// BEGIN MODIFIED BY Lkb - 8/jun/2001 - added support to compile without libpng
#ifndef LIBPNG_SUPPORT
	MessageBox(NULL, "Daedalus must be compiled with LIBPNG_SUPPORT defined to be able to take screenshots", "Error", 0);
	return;
#else
	HRESULT hr;
	
	LPDIRECT3DSURFACE8 pBackBuffer;
	
	hr = m_pd3dDevice->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
	if ( SUCCEEDED(hr) )
	{
		// Generate new filename
		LONG nCount;
		
		TCHAR szFilePath[MAX_PATH+1];
		TCHAR szFileName[MAX_PATH+1];
		TCHAR szFileExt[MAX_PATH+1];
		
		pBackBuffer->AddRef();
		Dump_GetDumpDirectory(szFilePath, TEXT("screenshots"));
		
		// Generate a filename that has not yet been used
		nCount = 0;
		do 
		{
			wsprintf(szFileExt, "%s%04d.png", g_ROM.settings.szGameName, nCount);
			PathCombine(szFileName, szFilePath, szFileExt);
			nCount++;
		} while ( PathFileExists( szFileName ) );
		
		DBGConsole_Msg(0, "Dumping screenshot as [C%s]", szFileName);	
		
		PngSaveSurface(szFileName, pBackBuffer);
		
		pBackBuffer->Release();
	}
#endif
	// END MODIFIED By Lkb - 8/jun/2001
}


//*****************************************************************************
//
//*****************************************************************************
HRESULT IGraphicsContext::GetBufferSize(DWORD * pdwWidth, DWORD * pdwHeight)
{
	*pdwWidth = m_d3dsdBackBuffer.Width;
	*pdwHeight = m_d3dsdBackBuffer.Height;
	
	return S_OK;
}

//*****************************************************************************
//
//*****************************************************************************
HRESULT IGraphicsContext::Initialise(HWND hWnd, HWND hWndStatus,
									 DWORD dwWidth, DWORD dwHeight,
									 BOOL bWindowed )
{
	HRESULT hr;
	
	Lock();
	
	// Create the Direct3D object
	m_pD3D = Direct3DCreate8( D3D_SDK_VERSION );
	if( m_pD3D == NULL )
	{
		Unlock();
		return DisplayErrorMsg( D3DAPPERR_NODIRECT3D, MSGERR_APPMUSTEXIT );
	}
	
	// Build a list of Direct3D adapters, modes and devices.
	if( FAILED( hr = BuildDeviceList() ) )
	{
		if ( m_pD3D )
		{
			m_pD3D->Release();
			m_pD3D = NULL;
		}
		Unlock();
		return DisplayErrorMsg( hr, MSGERR_APPMUSTEXIT );
	}
	
	
	m_hWnd = hWnd;
	m_hWndStatus = hWndStatus;
	
	
	// The focus window can be a specified to be a different window than the
	// device window.  If not, use the device window as the focus window.
	if( m_hWndFocus == NULL )
		m_hWndFocus = m_hWnd;
	
	// Save window properties
	m_dwWindowStyle = GetWindowLong( m_hWnd, GWL_STYLE );
	m_dwWindowExStyle = GetWindowLong( m_hWnd, GWL_EXSTYLE );
	m_dwStatusWindowStyle = GetWindowLong( m_hWndStatus, GWL_STYLE );
	
	RECT rcScreen, rcStatus;
	
	SetRect(&rcScreen, 0,0, dwWidth, dwHeight);
	
	// Calculate window size to give desired window size...
	AdjustWindowRectEx(&rcScreen, m_dwWindowStyle & (~(WS_THICKFRAME|WS_MAXIMIZEBOX)), TRUE, m_dwWindowExStyle);
	
	// Add extra margin for the status bar
	if ( IsWindow( m_hWndStatus ) )
	{
		// Add on enough space for the status bar
		GetClientRect(m_hWndStatus, &rcStatus);
		rcScreen.bottom += (rcStatus.bottom - rcStatus.top);
	}
	
	SetWindowPos(m_hWnd, 0,
		rcScreen.left, rcScreen.top,
		rcScreen.right-rcScreen.left, rcScreen.bottom-rcScreen.top,
		SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOZORDER);
	
	GetWindowRect( m_hWnd, &m_rcWindowBounds );
	GetClientRect( m_hWnd, &m_rcWindowClient );
	
	if ( IsWindow( m_hWndStatus ) )
	{
		// Remove statusbar area
		m_rcWindowClient.bottom -= (rcStatus.bottom - rcStatus.top);
	}
	
	
	hr = Create( bWindowed );
	
	
	// Clear/Update a few times to ensure that all of the buffers are cleared
	if ( m_pd3dDevice )
	{
		Clear(D3DCLEAR_ZBUFFER|D3DCLEAR_TARGET);
		m_pd3dDevice->BeginScene();
		m_pd3dDevice->EndScene();
		UpdateFrame();
		
		Clear(D3DCLEAR_ZBUFFER|D3DCLEAR_TARGET);
		m_pd3dDevice->BeginScene();
		m_pd3dDevice->EndScene();
		UpdateFrame();
	}
	
	
	Unlock();
	
	return hr;
}

//*****************************************************************************
// 
//*****************************************************************************
HRESULT IGraphicsContext::Create( BOOL bWindowed )
{
	HRESULT hr;
	
	// Get access to current adapter, device, and mode
	CDevice*  pDeviceInfo  = m_Adapters[m_dwAdapter].GetCurrentDevice();
	
	m_bWindowed = bWindowed;		// Make user Toggle manually for now!
	pDeviceInfo->SetWindowed( m_bWindowed ? true : false );
	
	// Initialize the 3D environment for the app
	if( FAILED( hr = Initialize3DEnvironment() ) )
	{
		SAFE_RELEASE( m_pD3D );
		return DisplayErrorMsg( hr, MSGERR_APPMUSTEXIT );
	}
	
	// The app is ready to go
	m_bReady = TRUE;
	
	return S_OK;
}


//-----------------------------------------------------------------------------
// Name: Initialize3DEnvironment()
// Desc:
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::Initialize3DEnvironment()
{
	HRESULT hr;
	
	const CAdapter* pAdapterInfo = &m_Adapters[m_dwAdapter];
	const CDevice*  pDeviceInfo  = m_Adapters[m_dwAdapter].GetCurrentDevice();
	const CMode*    pModeInfo    = pDeviceInfo->GetCurrentMode();
	
	// Prepare window for possible windowed/fullscreen change
	AdjustWindowForChange();
	
	// Set up the presentation parameters
	ZeroMemory( &m_d3dpp, sizeof(m_d3dpp) );
	m_d3dpp.Windowed               = pDeviceInfo->GetWindowed();
	m_d3dpp.BackBufferCount        = 1;
	m_d3dpp.MultiSampleType        = pDeviceInfo->GetMultiSampleType();
	m_d3dpp.SwapEffect             = D3DSWAPEFFECT_DISCARD;
	m_d3dpp.EnableAutoDepthStencil = TRUE;
	m_d3dpp.AutoDepthStencilFormat = pModeInfo->DepthStencilFormat;
	m_d3dpp.hDeviceWindow          = m_hWnd;
	m_d3dpp.Flags				   = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
	if( m_bWindowed )
	{
		m_d3dpp.BackBufferWidth  = m_rcWindowClient.right - m_rcWindowClient.left;
		m_d3dpp.BackBufferHeight = m_rcWindowClient.bottom - m_rcWindowClient.top;
		m_d3dpp.BackBufferFormat = pAdapterInfo->GetDesktopDisplayMode().Format;
	}
	else
	{
		m_d3dpp.BackBufferWidth  = pModeInfo->Width;
		m_d3dpp.BackBufferHeight = pModeInfo->Height;
		m_d3dpp.BackBufferFormat = pModeInfo->Format;
	}
	
	// Create the device
	hr = m_pD3D->CreateDevice( m_dwAdapter, pDeviceInfo->GetType(),
								m_hWndFocus, pModeInfo->dwBehavior, &m_d3dpp,
								&m_pd3dDevice );
	if( SUCCEEDED(hr) )
	{
		g_pD3DDev = m_pd3dDevice;
		
		// When moving from fullscreen to windowed mode, it is important to
		// adjust the window size after recreating the device rather than
		// beforehand to ensure that you get the window size you want.  For
		// example, when switching from 640x480 fullscreen to windowed with
		// a 1000x600 window on a 1024x768 desktop, it is impossible to set
		// the window size to 1000x600 until after the display mode has
		// changed to 1024x768, because windows cannot be larger than the
		// desktop.
		if( m_bWindowed )
		{
			SetWindowPos( m_hWnd, HWND_NOTOPMOST,
				m_rcWindowBounds.left, m_rcWindowBounds.top,
				( m_rcWindowBounds.right - m_rcWindowBounds.left ),
				( m_rcWindowBounds.bottom - m_rcWindowBounds.top ),
				SWP_SHOWWINDOW );
		}
		
		// Store device Caps
		m_pd3dDevice->GetDeviceCaps( &m_d3dCaps );
		
		// Store device description
		if( pDeviceInfo->GetType() == D3DDEVTYPE_REF )
			lstrcpy( m_strDeviceStats, "REF" );
		else if( pDeviceInfo->GetType() == D3DDEVTYPE_HAL )
			lstrcpy( m_strDeviceStats, "HAL" );
		else if( pDeviceInfo->GetType() == D3DDEVTYPE_SW )
			lstrcpy( m_strDeviceStats, "SW" );
		
		if( pModeInfo->dwBehavior & D3DCREATE_HARDWARE_VERTEXPROCESSING &&
			pModeInfo->dwBehavior & D3DCREATE_PUREDEVICE )
		{
			if( pDeviceInfo->GetType() == D3DDEVTYPE_HAL )
				lstrcat( m_strDeviceStats, " (pure hw vp)" );
			else
				lstrcat( m_strDeviceStats, " (simulated pure hw vp)" );
		}
		else if( pModeInfo->dwBehavior & D3DCREATE_HARDWARE_VERTEXPROCESSING )
		{
			if( pDeviceInfo->GetType() == D3DDEVTYPE_HAL )
				lstrcat( m_strDeviceStats, " (hw vp)" );
			else
				lstrcat( m_strDeviceStats, " (simulated hw vp)" );
		}
		else if( pModeInfo->dwBehavior & D3DCREATE_MIXED_VERTEXPROCESSING )
		{
			if( pDeviceInfo->GetType() == D3DDEVTYPE_HAL )
				lstrcat( m_strDeviceStats, " (mixed vp)" );
			else
				lstrcat( m_strDeviceStats, " (simulated mixed vp)" );
		}
		else if( pModeInfo->dwBehavior & D3DCREATE_SOFTWARE_VERTEXPROCESSING )
		{
			lstrcat( m_strDeviceStats, " (sw vp)" );
		}
		
		if( pDeviceInfo->GetType() == D3DDEVTYPE_HAL )
		{
			lstrcat( m_strDeviceStats, ": " );
			lstrcat( m_strDeviceStats, pAdapterInfo->GetAdapterID().Description );
		}
		
		if ( IsWindow(m_hWndStatus ) )
		{
			::SendMessage(m_hWndStatus, SB_SETTEXT, 0, (LPARAM)m_strDeviceStats);
		}
		
		// Store render target surface desc
		LPDIRECT3DSURFACE8 pBackBuffer;
		m_pd3dDevice->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
		pBackBuffer->GetDesc( &m_d3dsdBackBuffer );
		pBackBuffer->Release();
		
		// Set up the fullscreen cursor
		if( !m_bWindowed )
		{
			m_pd3dDevice->ShowCursor( FALSE );
		}
		
		// Initialize the app's device-dependent objects
		hr = InitDeviceObjects();
		if( SUCCEEDED(hr) )
		{
			hr = RestoreDeviceObjects();
			if( SUCCEEDED(hr) )
			{
				m_bActive = TRUE;
				return S_OK;
			}
		}
		
		// Cleanup before we try again (- shouldn't get here)
		InvalidateDeviceObjects();
		DeleteDeviceObjects();
		SAFE_RELEASE( m_pd3dDevice );
	}
	
	// If that failed, fall back to the reference rasterizer (removed)
	
	return hr;
}


//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::Resize3DEnvironment()
{
	HRESULT hr;
	
	// Release all vidmem objects
	if( FAILED( hr = InvalidateDeviceObjects() ) )
		return hr;
	
	// Reset the device
	if( FAILED( hr = m_pd3dDevice->Reset( &m_d3dpp ) ) )
	{
		while ( hr == D3DERR_DEVICELOST )
		{
			Pause(1);
			hr = m_pd3dDevice->Reset( &m_d3dpp );
		}
	}
	
	if ( FAILED(hr) )
		return hr;
	
	// Store render target surface desc
	LPDIRECT3DSURFACE8 pBackBuffer;
	m_pd3dDevice->GetBackBuffer( 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
	pBackBuffer->GetDesc( &m_d3dsdBackBuffer );
	pBackBuffer->Release();
	
	// Set up the fullscreen cursor
	if( !m_bWindowed )
	{
		m_pd3dDevice->ShowCursor( FALSE );
	}
	
	
	// Initialize the app's device-dependent objects
	hr = RestoreDeviceObjects();
	if( FAILED(hr) )
		return hr;
	
	return S_OK;
}


//-----------------------------------------------------------------------------
// Name: ToggleFullScreen()
// Desc: Called when user toggles between fullscreen mode and windowed mode
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::DoToggleFullscreen()
{
	// Get access to current adapter, device, and mode
	CAdapter* pAdapterInfo = &m_Adapters[m_dwAdapter];
	CDevice*  pDeviceInfo  = m_Adapters[m_dwAdapter].GetCurrentDevice();
	CMode*	  pModeInfo    = pDeviceInfo->GetCurrentMode();
	
	// Need device change if going windowed and the current device
	// can only be fullscreen
	if( !m_bWindowed && !pDeviceInfo->CanDoWindowed() )
		return ForceWindowed();
	
	m_bReady = FALSE;
	
	// Toggle the windowed state
	m_bWindowed = !m_bWindowed;
	pDeviceInfo->SetWindowed( m_bWindowed ? true : false );
	
	// Prepare window for windowed/fullscreen change
	AdjustWindowForChange();
	
	// Set up the presentation parameters
	m_d3dpp.Windowed			   = pDeviceInfo->GetWindowed();
	m_d3dpp.MultiSampleType 	   = pDeviceInfo->GetMultiSampleType();
	m_d3dpp.AutoDepthStencilFormat = pModeInfo->DepthStencilFormat;
	m_d3dpp.hDeviceWindow		   = m_hWnd;
	if( m_bWindowed )
	{
		m_d3dpp.BackBufferWidth  = m_rcWindowClient.right - m_rcWindowClient.left;
		m_d3dpp.BackBufferHeight = m_rcWindowClient.bottom - m_rcWindowClient.top;
		m_d3dpp.BackBufferFormat = pAdapterInfo->GetDesktopDisplayMode().Format;
	}
	else
	{
		m_d3dpp.BackBufferWidth  = pModeInfo->Width;
		m_d3dpp.BackBufferHeight = pModeInfo->Height;
		m_d3dpp.BackBufferFormat = pModeInfo->Format;
	}
	
	// Resize the 3D device
	if( FAILED( Resize3DEnvironment() ) )
	{
		if( m_bWindowed )
			return ForceWindowed();
		else
			return E_FAIL;
	}
	
	// When moving from fullscreen to windowed mode, it is important to
	// adjust the window size after resetting the device rather than
	// beforehand to ensure that you get the window size you want.	For
	// example, when switching from 640x480 fullscreen to windowed with
	// a 1000x600 window on a 1024x768 desktop, it is impossible to set
	// the window size to 1000x600 until after the display mode has
	// changed to 1024x768, because windows cannot be larger than the
	// desktop.
	if( m_bWindowed )
	{
		SetWindowPos( m_hWnd, HWND_NOTOPMOST,
			m_rcWindowBounds.left, m_rcWindowBounds.top,
			( m_rcWindowBounds.right - m_rcWindowBounds.left ),
			( m_rcWindowBounds.bottom - m_rcWindowBounds.top ),
			SWP_SHOWWINDOW );
	}
	
	m_bReady = TRUE;
	
	return S_OK;
}




//-----------------------------------------------------------------------------
// Name: ForceWindowed()
// Desc: Switch to a windowed mode, even if that means picking a new device
//		 and/or adapter
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::ForceWindowed()
{
	HRESULT hr;
	CAdapter* pAdapterInfoCur = &m_Adapters[m_dwAdapter];
	CDevice*  pDeviceInfoCur  = m_Adapters[m_dwAdapter].GetCurrentDevice();
	BOOL bFoundDevice = FALSE;
	
	if( pDeviceInfoCur->CanDoWindowed() )
	{
		bFoundDevice = TRUE;
	}
	else
	{
		// Look for a windowable device on any adapter
		CAdapter* pAdapterInfo;
		DWORD dwAdapter;
		CDevice* pDeviceInfo;
		DWORD dwDevice;
		for( dwAdapter = 0; dwAdapter < m_dwNumAdapters; dwAdapter++ )
		{
			pAdapterInfo = &m_Adapters[dwAdapter];
			for( dwDevice = 0; dwDevice < pAdapterInfo->GetNumDevices(); dwDevice++ )
			{
				pDeviceInfo = pAdapterInfo->GetDevice( dwDevice );
				if( pDeviceInfo->CanDoWindowed() )
				{
					m_dwAdapter = dwAdapter;
					pDeviceInfoCur = pDeviceInfo;
					pAdapterInfo->SetCurrentDevice( dwDevice );
					bFoundDevice = TRUE;
					break;
				}
			}
			if( bFoundDevice )
				break;
		}
	}
	
	if( !bFoundDevice )
		return E_FAIL;
	
	pDeviceInfoCur->SetWindowed( true );
	m_bWindowed = TRUE;
	
	// Now destroy the current 3D device objects, then reinitialize
	
	m_bReady = FALSE;
	
	// Release all scene objects that will be re-created for the new device
	InvalidateDeviceObjects();
	DeleteDeviceObjects();
	
	// Release display objects, so a new device can be created
	LONG nRefCount = m_pd3dDevice->Release();
	if( nRefCount > 0L )
	{
		DBGConsole_Msg(0, "Refcount of device is %d", nRefCount );
		return DisplayErrorMsg( D3DAPPERR_NONZEROREFCOUNT, MSGERR_APPMUSTEXIT );
	}
	
	// Create the new device
	if( FAILED( hr = Initialize3DEnvironment() ) )
		return DisplayErrorMsg( hr, MSGERR_APPMUSTEXIT );
	m_bReady = TRUE;
	
	return S_OK;
}



//-----------------------------------------------------------------------------
// Name: AdjustWindowForChange()
// Desc: Prepare the window for a possible change between windowed mode and
//		 fullscreen mode.  This function is virtual and thus can be overridden
//		 to provide different behavior, such as switching to an entirely
//		 different window for fullscreen mode (as in the MFC sample apps).
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::AdjustWindowForChange()
{
	if( m_bWindowed )
	{
		// Set windowed-mode style - but disable resizing
		SetWindowLong( m_hWnd, GWL_STYLE, m_dwWindowStyle & (~(WS_THICKFRAME|WS_MAXIMIZEBOX)) );
		
		if ( IsWindow(m_hWndStatus) )
		{
			SetWindowLong( m_hWndStatus, GWL_STYLE, m_dwStatusWindowStyle & (~SBARS_SIZEGRIP));
		}
	}
	else
	{
		// Set fullscreen-mode style
		SetWindowLong( m_hWnd, GWL_STYLE, WS_POPUP|WS_SYSMENU|WS_VISIBLE );
	}
	return S_OK;
}





//-----------------------------------------------------------------------------
// Name: UserSelectNewDevice()
// Desc: Displays a dialog so the user can select a new adapter, device, or
//		 display mode, and then recreates the 3D environment if needed
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::UserSelectNewDevice()
{
	HRESULT hr;
	
	// Can't display dialogs in fullscreen mode
	if( m_bWindowed == FALSE )
	{
		if( FAILED( DoToggleFullscreen() ) )
		{
			DisplayErrorMsg( D3DAPPERR_RESIZEFAILED, MSGERR_APPMUSTEXIT );
			return E_FAIL;
		}
	}
	
	// Prompt the user to change the mode
	if( IDOK != DialogBoxParam( (HINSTANCE)GetModuleHandle(NULL),
		MAKEINTRESOURCE(IDD_SELECTDEVICE), m_hWnd,
		SelectDeviceProc, (LPARAM)this ) )
		return S_OK;
	
	// Get access to the newly selected adapter, device, and mode
	m_bWindowed = m_Adapters[m_dwAdapter].GetCurrentDevice()->GetWindowed();
	
	// Release all scene objects that will be re-created for the new device
	InvalidateDeviceObjects();
	DeleteDeviceObjects();
	
	// Release display objects, so a new device can be created
	LONG nRefCount = m_pd3dDevice->Release();
	if( nRefCount > 0L )
	{
		DBGConsole_Msg(0, "Refcount of device is %d", nRefCount );
		return DisplayErrorMsg( D3DAPPERR_NONZEROREFCOUNT, MSGERR_APPMUSTEXIT );
	}
	
	// Inform the display class of the change. It will internally
	// re-create valid surfaces, a d3ddevice, etc.
	if( FAILED( hr = Initialize3DEnvironment() ) )
		return DisplayErrorMsg( hr, MSGERR_APPMUSTEXIT );
	
	return S_OK;
}




//-----------------------------------------------------------------------------
// Name: SelectDeviceProc()
// Desc: Windows message handling function for the device select dialog
//-----------------------------------------------------------------------------
INT_PTR CALLBACK IGraphicsContext::SelectDeviceProc( HWND hDlg, UINT msg,
													WPARAM wParam, LPARAM lParam )
{
	// Get access to the UI controls
	HWND hwndAdapterList		= GetDlgItem( hDlg, IDC_ADAPTER_COMBO );
	HWND hwndDeviceList 		= GetDlgItem( hDlg, IDC_DEVICE_COMBO );
	HWND hwndFullscreenModeList = GetDlgItem( hDlg, IDC_FULLSCREENMODES_COMBO );
	HWND hwndWindowedRadio		= GetDlgItem( hDlg, IDC_WINDOW );
	HWND hwndFullscreenRadio	= GetDlgItem( hDlg, IDC_FULLSCREEN );
	HWND hwndMultiSampleList	= GetDlgItem( hDlg, IDC_MULTISAMPLE_COMBO );
	BOOL bUpdateDlgControls 	= FALSE;
	
	// Static state for adapter/device/mode selection
	static IGraphicsContext * pd3dCtx;
	static DWORD  dwOldAdapter, dwNewAdapter;
	static DWORD  dwOldDevice,	dwNewDevice;
	static DWORD  dwOldMode,	dwNewMode;
	static bool   bOldWindowed, bNewWindowed;
	static D3DMULTISAMPLE_TYPE OldMultiSampleType, NewMultiSampleType;
	
	// Working variables
	CAdapter* pAdapter;
	CDevice*  pDevice;
	
	// Handle the initialization message
	if( WM_INITDIALOG == msg )
	{
		// Old state
		pd3dCtx 	 = (IGraphicsContext*)lParam;
		dwOldAdapter = pd3dCtx->m_dwAdapter;
		pAdapter	 = &pd3dCtx->m_Adapters[dwOldAdapter];
		
		dwOldDevice  = pAdapter->GetCurrentDeviceIdx();
		pDevice 	 = pAdapter->GetCurrentDevice();
		
		dwOldMode	 = pDevice->GetCurrentModeIdx();
		bOldWindowed = pDevice->GetWindowed();
		OldMultiSampleType = pDevice->GetMultiSampleType();
		
		// New state is initially the same as the old state
		dwNewAdapter = dwOldAdapter;
		dwNewDevice  = dwOldDevice;
		dwNewMode	 = dwOldMode;
		bNewWindowed = bOldWindowed;
		NewMultiSampleType = OldMultiSampleType;
		
		// Set flag to update dialog controls below
		bUpdateDlgControls = TRUE;
	}
	
	if( WM_COMMAND == msg )
	{
		// Get current UI state
		bNewWindowed  = Button_GetCheck( hwndWindowedRadio ) ? true : false;
		
		if( IDOK == LOWORD(wParam) )
		{
			// Handle the case when the user hits the OK button. Check if any
			// of the options were changed
			if( dwNewAdapter != dwOldAdapter || dwNewDevice  != dwOldDevice  ||
				dwNewMode	 != dwOldMode	 || bNewWindowed != bOldWindowed ||
				NewMultiSampleType != OldMultiSampleType )
			{
				pd3dCtx->m_dwAdapter = dwNewAdapter;
				
				pAdapter = &pd3dCtx->m_Adapters[dwNewAdapter];
				pAdapter->SetCurrentDevice( dwNewDevice );
				
				CDevice * pDevice = pAdapter->GetCurrentDevice();
				
				pDevice->SetCurrentMode( dwNewMode );
				pDevice->SetWindowed( bNewWindowed );
				pDevice->SetMultiSampleType( NewMultiSampleType );
				
				EndDialog( hDlg, IDOK );
			}
			else
				EndDialog( hDlg, IDCANCEL );
			
			return TRUE;
		}
		else if( IDCANCEL == LOWORD(wParam) )
		{
			// Handle the case when the user hits the Cancel button
			EndDialog( hDlg, IDCANCEL );
			return TRUE;
		}
		else if( CBN_SELENDOK == HIWORD(wParam) )
		{
			if( LOWORD(wParam) == IDC_ADAPTER_COMBO )
			{
				dwNewAdapter = ComboBox_GetCurSel( hwndAdapterList );
				pAdapter	 = &pd3dCtx->m_Adapters[dwNewAdapter];
				
				dwNewDevice  = pAdapter->GetCurrentDeviceIdx();
				dwNewMode	 = pAdapter->GetDevice( dwNewDevice )->GetCurrentModeIdx();
				bNewWindowed = pAdapter->GetDevice( dwNewDevice )->GetWindowed();
			}
			else if( LOWORD(wParam) == IDC_DEVICE_COMBO )
			{
				pAdapter	 = &pd3dCtx->m_Adapters[dwNewAdapter];
				
				dwNewDevice  = ComboBox_GetCurSel( hwndDeviceList );
				dwNewMode	 = pAdapter->GetDevice( dwNewDevice )->GetCurrentModeIdx();
				bNewWindowed = pAdapter->GetDevice( dwNewDevice )->GetWindowed();
			}
			else if( LOWORD(wParam) == IDC_FULLSCREENMODES_COMBO )
			{
				dwNewMode = ComboBox_GetCurSel( hwndFullscreenModeList );
			}
			else if( LOWORD(wParam) == IDC_MULTISAMPLE_COMBO )
			{
				DWORD dwItem = ComboBox_GetCurSel( hwndMultiSampleList );
				NewMultiSampleType = (D3DMULTISAMPLE_TYPE)ComboBox_GetItemData( hwndMultiSampleList, dwItem );
			}
		}
		// Keep the UI current
		bUpdateDlgControls = TRUE;
	}
	
	// Update the dialog controls
	if( bUpdateDlgControls )
	{
		// Reset the content in each of the combo boxes
		ComboBox_ResetContent( hwndAdapterList );
		ComboBox_ResetContent( hwndDeviceList );
		ComboBox_ResetContent( hwndFullscreenModeList );
		ComboBox_ResetContent( hwndMultiSampleList );
		
		pAdapter = &pd3dCtx->m_Adapters[dwNewAdapter];
		pDevice  = pAdapter->GetDevice( dwNewDevice );
		
		// Add a list of adapters to the adapter combo box
		for( DWORD a=0; a < pd3dCtx->m_dwNumAdapters; a++ )
		{
			// Add device name to the combo box
			DWORD dwItem = ComboBox_AddString( hwndAdapterList,
				pd3dCtx->m_Adapters[a].GetAdapterID().Description );
			
			// Set the item data to identify this adapter
			ComboBox_SetItemData( hwndAdapterList, dwItem, a );
			
			// Set the combobox selection on the current adapater
			if( a == dwNewAdapter )
				ComboBox_SetCurSel( hwndAdapterList, dwItem );
		}
		
		// Add a list of devices to the device combo box
		for( DWORD d=0; d < pAdapter->GetNumDevices(); d++ )
		{
			// Add device name to the combo box
			DWORD dwItem = ComboBox_AddString( hwndDeviceList, pAdapter->GetDevice( d )->GetDescription() );
			
			// Set the item data to identify this device
			ComboBox_SetItemData( hwndDeviceList, dwItem, d );
			
			// Set the combobox selection on the current device
			if( d == dwNewDevice )
				ComboBox_SetCurSel( hwndDeviceList, dwItem );
		}
		
		// Add a list of modes to the mode combo box
		for( DWORD m=0; m < pDevice->GetNumModes(); m++ )
		{
			DWORD BitDepth = 16;
			if( pDevice->GetMode( m )->Format == D3DFMT_X8R8G8B8 ||
				pDevice->GetMode( m )->Format == D3DFMT_A8R8G8B8 ||
				pDevice->GetMode( m )->Format == D3DFMT_R8G8B8 )
			{
				BitDepth = 32;
			}
			
			// Add mode desc to the combo box
			TCHAR strMode[80];
			sprintf( strMode, "%ld x %ld x %ld", pDevice->GetMode( m )->Width, pDevice->GetMode( m )->Height, BitDepth );
			DWORD dwItem = ComboBox_AddString( hwndFullscreenModeList, strMode );
			
			// Set the item data to identify this mode
			ComboBox_SetItemData( hwndFullscreenModeList, dwItem, m );
			
			// Set the combobox selection on the current mode
			if( m == dwNewMode )
				ComboBox_SetCurSel( hwndFullscreenModeList, dwItem );
		}
		
		// Add a list of multisample modes to the multisample combo box
		for( m=0; m <= 16; m++ )
		{
			TCHAR strDesc[50];
			
			D3DFORMAT fmt;
			if( bNewWindowed )
				fmt = pd3dCtx->m_Adapters[dwNewAdapter].GetDesktopDisplayMode().Format;
			else
				fmt = pDevice->GetMode( dwNewMode )->Format;
			
			if ( m == 1 ) // 1 is not a valid multisample type
				continue;
			
			if( SUCCEEDED( pd3dCtx->m_pD3D->CheckDeviceMultiSampleType( dwNewAdapter,
				pDevice->GetType(), fmt, bNewWindowed, (D3DMULTISAMPLE_TYPE)m ) ) )
			{
				if( m == 0 )
					lstrcpy( strDesc, "none" );
				else
					wsprintf( strDesc, "%d samples", m );
				
				// Add device name to the combo box
				DWORD dwItem = ComboBox_AddString( hwndMultiSampleList, strDesc );
				
				// Set the item data to identify this multisample type
				ComboBox_SetItemData( hwndMultiSampleList, dwItem, m );
				
				// Set the combobox selection on the current multisample type
				if( (D3DMULTISAMPLE_TYPE)m == NewMultiSampleType || m == 0 )
					ComboBox_SetCurSel( hwndMultiSampleList, dwItem );
			}
		}
		DWORD dwItem = ComboBox_GetCurSel( hwndMultiSampleList );
		NewMultiSampleType = (D3DMULTISAMPLE_TYPE)ComboBox_GetItemData( hwndMultiSampleList, dwItem );
		EnableWindow( hwndMultiSampleList, ComboBox_GetCount( hwndMultiSampleList ) > 1);
		EnableWindow( hwndWindowedRadio, pDevice->CanDoWindowed() );
		
		if( bNewWindowed )
		{
			Button_SetCheck( hwndWindowedRadio,   TRUE );
			Button_SetCheck( hwndFullscreenRadio, FALSE );
			EnableWindow( hwndFullscreenModeList, FALSE );
		}
		else
		{
			Button_SetCheck( hwndWindowedRadio,   FALSE );
			Button_SetCheck( hwndFullscreenRadio, TRUE );
			EnableWindow( hwndFullscreenModeList, TRUE );
		}
		return TRUE;
	}
	
	return FALSE;
}



//-----------------------------------------------------------------------------
// Name: BuildDeviceList()
// Desc: From DX8 SDK Copyright (c) 1998-2000 Microsoft
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::BuildDeviceList()
{
	BuildAdapterList();
	
	// Return an error if no compatible devices were found
	if( 0L == m_dwNumAdapters )
		return D3DAPPERR_NOCOMPATIBLEDEVICES;
	
	// Pick a default device that can render into a window
	// (This code assumes that the HAL device comes before the REF
	// device in the device array).
	for( DWORD a=0; a<m_dwNumAdapters; a++ )
	{
		for( DWORD d=0; d < m_Adapters[a].GetNumDevices(); d++ )
		{
			if( m_Adapters[a].GetDevice( d )->GetWindowed() )
			{
				m_Adapters[a].SetCurrentDevice( d );
				m_dwAdapter = a;
				m_bWindowed = TRUE;
				
				// Display a warning message
				if( m_Adapters[a].GetDevice( d )->GetType() == D3DDEVTYPE_REF )
				{
					DisplayErrorMsg( D3DAPPERR_NOHALTHISMODE, MSGWARN_SWITCHEDTOREF );
				}
				
				return S_OK;
			}
		}
	}
	
	return D3DAPPERR_NOWINDOWABLEDEVICES;
}


//*****************************************************************************
// 
//*****************************************************************************
void IGraphicsContext::BuildAdapterList( )
{
	// Loop through all the adapters on the system (usually, there's just one
	// unless more than one graphics card is present).
	for( UINT iAdapter = 0; iAdapter < m_pD3D->GetAdapterCount(); iAdapter++ )
	{
		// Fill in adapter info
		CAdapter adapter;
		
		adapter.Initialise( m_pD3D, iAdapter );
	
		// If valid devices were found, keep this adapter
		if( adapter.GetNumDevices() > 0 )
		{
			m_Adapters[ m_dwNumAdapters++ ] = adapter;
		}
	}
}


//-----------------------------------------------------------------------------
// Name: DisplayErrorMsg()
// Desc: Displays error messages in a message box
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::DisplayErrorMsg( HRESULT hr, DWORD dwType )
{
	TCHAR strMsg[512];
	
	switch( hr )
	{
	case D3DAPPERR_NODIRECT3D:
		strcpy( strMsg, "Could not initialize Direct3D. You may\n"
						"want to check that the latest version of\n"
						"DirectX is correctly installed on your\n"
						"system." );
		break;
		
	case D3DAPPERR_NOCOMPATIBLEDEVICES:
		strcpy( strMsg, "Could not find any compatible Direct3D\n"
						"devices." );
		break;
		
	case D3DAPPERR_NOWINDOWABLEDEVICES:
		strcpy( strMsg, "This program cannot run in a desktop\n"
						"window with the current display settings.\n"
						"Please change your desktop settings to a\n"
						"16- or 32-bit display mode and re-run this\n"
						"sample." );
		break;
							
	case D3DAPPERR_NOHALTHISMODE:
		strcpy( strMsg, "This program requires functionality that is\n"
						"not available on your Direct3D hardware\n"
						"accelerator with the current desktop display\n"
						"settings." );
		break;
		
	case D3DAPPERR_RESIZEFAILED:
		strcpy( strMsg, "Could not reset the Direct3D device." );
		break;
		
	case D3DAPPERR_NONZEROREFCOUNT:
		strcpy( strMsg, "A D3D object has a non-zero reference\n"
						"count (meaning things were not properly\n"
						"cleaned up)." );
		break;
		
	case E_OUTOFMEMORY:
		strcpy( strMsg, "Not enough memory." );
		break;
		
	case D3DERR_OUTOFVIDEOMEMORY:
		strcpy( strMsg, "Not enough video memory." );
		break;
		
	default:
		strcpy( strMsg, "Generic application error. Enable\n"
						"debug output for detailed information." );
	}
	
	// TODO: Use IDS_D3DERROR resource
	// IDS_D3DERROR has one %s for the error message
	//wsprintf(szMsg, CResourceString(IDS_D3DERROR), szError);
	
	
	if( MSGERR_APPMUSTEXIT == dwType )
	{
		strcat( strMsg, "\n\nThis program will now exit." );
		//MessageBox( NULL, strMsg, g_szDaedalusName, MB_ICONERROR|MB_OK );
		
		// Close the window, which shuts down the app
		//if( m_hWnd )
		//	  SendMessage( m_hWnd, WM_CLOSE, 0, 0 );
	}
	else
	{
		if( MSGWARN_SWITCHEDTOREF == dwType )
			strcat( strMsg, "\n\nSwitching to the reference rasterizer,\n"
							"a software device that implements the entire\n"
							"Direct3D feature set, but runs very slowly." );
		MessageBox( NULL, strMsg, g_szDaedalusName, MB_ICONWARNING|MB_OK );
	}
	
	OutputDebugString( strMsg );
	
	return hr;
}






//-----------------------------------------------------------------------------
// Name: Cleanup3DEnvironment()
// Desc: Cleanup scene objects
//-----------------------------------------------------------------------------
VOID IGraphicsContext::Cleanup3DEnvironment()
{
	m_bActive = FALSE;
	m_bReady  = FALSE;
	
	if ( IsWindow( m_hWnd ) )
	{
		SetWindowLong( m_hWnd, GWL_STYLE, m_dwWindowStyle );
	}
	
	if ( IsWindow(m_hWndStatus) )
	{
		SetWindowLong( m_hWndStatus, GWL_STYLE, m_dwStatusWindowStyle);
	}
	
	if( m_pd3dDevice )
	{
		InvalidateDeviceObjects();
		DeleteDeviceObjects();
		
		m_pd3dDevice->Release();
		m_pD3D->Release();
		
		m_pd3dDevice = NULL;
		m_pD3D		 = NULL;
	}
	
	// FinalCleanup();
}


LRESULT IGraphicsContext::OnSize( WPARAM state )
{
	// Check to see if we are losing our window...
	if( SIZE_MAXHIDE==state || SIZE_MINIMIZED==state )
		m_bActive = FALSE;
	else
		m_bActive = TRUE;
	
	return 0;//DefWindowProc( m_hWnd, uMsg, wParam, lParam );
}

LRESULT IGraphicsContext::OnExitSizeMove()
{
	HRESULT hr;
	if( m_bActive && m_bWindowed )
	{
		RECT rcClientOld;
		rcClientOld = m_rcWindowClient;
		
		// Update window properties
		GetWindowRect( m_hWnd, &m_rcWindowBounds );
		GetClientRect( m_hWnd, &m_rcWindowClient );
		
		if( rcClientOld.right - rcClientOld.left !=
			m_rcWindowClient.right - m_rcWindowClient.left ||
			rcClientOld.bottom - rcClientOld.top !=
			m_rcWindowClient.bottom - m_rcWindowClient.top)
		{
			// A new window size will require a new backbuffer
			// size, so the 3D structures must be changed accordingly.
			m_bReady = FALSE;
			
			m_d3dpp.BackBufferWidth  = m_rcWindowClient.right - m_rcWindowClient.left;
			m_d3dpp.BackBufferHeight = m_rcWindowClient.bottom - m_rcWindowClient.top;
			
			// Resize the 3D environment
			if( FAILED( hr = Resize3DEnvironment() ) )
			{
				DisplayErrorMsg( D3DAPPERR_RESIZEFAILED, MSGERR_APPMUSTEXIT );
				return 0;
			}
			
			m_bReady = TRUE;
		}
	}
	
	return 0;//DefWindowProc( hWnd, uMsg, wParam, lParam );
}

void IGraphicsContext::OnChangeDevice()
{
	HRESULT hr;
	// Prompt the user to select a new device or mode
	if( m_bActive && m_bReady )
	{
		Pause(TRUE);
		
		if( FAILED( hr = UserSelectNewDevice() ) )
			return;
		
		Pause(FALSE);
	}
}

void IGraphicsContext::ToggleFullscreen()
{
	// Toggle the fullscreen/window mode
	if( m_bActive && m_bReady )
	{
		if( FAILED( DoToggleFullscreen() ) )
		{
			DisplayErrorMsg( D3DAPPERR_RESIZEFAILED, MSGERR_APPMUSTEXIT );
			return;
		}
	}
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message handling function.
//-----------------------------------------------------------------------------
LRESULT IGraphicsContext::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	case WM_PAINT:
		// Handle paint messages when the app is not ready
		if( m_pd3dDevice && !m_bReady )
		{
			if( m_bWindowed )
				m_pd3dDevice->Present( NULL, NULL, NULL, NULL );
		}
		break;
		
	case WM_GETMINMAXINFO:
		((MINMAXINFO*)lParam)->ptMinTrackSize.x = 100;
		((MINMAXINFO*)lParam)->ptMinTrackSize.y = 100;
		break;
		
	case WM_ENTERSIZEMOVE:
		break;
		
		
	case WM_SETCURSOR:
		// Turn off Windows cursor in fullscreen mode
		if( m_bActive && m_bReady && !m_bWindowed )
		{
			SetCursor( NULL );
			return TRUE; // prevent Windows from setting cursor to window class cursor
		}
		break;
		
	case WM_MOUSEMOVE:
		if( m_bActive && m_bReady && m_pd3dDevice != NULL )
		{
			POINT ptCursor;
			GetCursorPos( &ptCursor );
			ScreenToClient( m_hWnd, &ptCursor );
			m_pd3dDevice->SetCursorPosition( ptCursor.x, ptCursor.y, 0L );
		}
		break;
		
	case WM_ENTERMENULOOP:
		// Pause the app when menus are displayed
		Pause(TRUE);
		break;
		
	case WM_EXITMENULOOP:
		Pause(FALSE);
		break;
		
	case WM_NCHITTEST:
		// Prevent the user from selecting the menu in fullscreen mode
		if( !m_bWindowed )
			return HTCLIENT;
		
		break;
		
	case WM_POWERBROADCAST:
		switch( wParam )
		{
#ifndef PBT_APMQUERYSUSPEND
#define PBT_APMQUERYSUSPEND 0x0000
#endif
		case PBT_APMQUERYSUSPEND:
			// At this point, the app should save any data for open
			// network connections, files, etc., and prepare to go into
			// a suspended mode.
			return TRUE;
			
#ifndef PBT_APMRESUMESUSPEND
#define PBT_APMRESUMESUSPEND 0x0007
#endif
		case PBT_APMRESUMESUSPEND:
			// At this point, the app should recover any data, network
			// connections, files, etc., and resume running from when
			// the app was suspended.
			return TRUE;
		}
		break;
		
		case WM_SYSCOMMAND:
			// Prevent moving/sizing and power loss in fullscreen mode
			switch( wParam )
			{
			case SC_MOVE:
			case SC_SIZE:
			case SC_MAXIMIZE:
			case SC_KEYMENU:
			case SC_MONITORPOWER:
				if( FALSE == m_bWindowed )
					return 1;
				break;
			}
			break;
			
			case WM_CLOSE:
				Cleanup3DEnvironment();
				DestroyMenu( GetMenu(hWnd) );
				return 0;
	}
	
	return DefWindowProc( hWnd, uMsg, wParam, lParam );
}






//-----------------------------------------------------------------------------
// Name: Pause()
// Desc: Called in to toggle the pause state of the app.
//-----------------------------------------------------------------------------
VOID IGraphicsContext::Pause( BOOL bPause )
{
	static DWORD dwAppPausedCount = 0L;
	
	dwAppPausedCount += ( bPause ? +1 : -1 );
	m_bReady		  = ( dwAppPausedCount ? FALSE : TRUE );
}





//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize device-dependent objects. This is the place to create mesh
//		 and texture objects.
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::InitDeviceObjects()
{
	// Set the device's states for rendering
	//return RDPInit();
	
	return S_OK;
}



//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//		 resized.
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::RestoreDeviceObjects()
{
	if ( D3DRender::IsAvailable() )
	{
		D3DRender::Get()->RestoreDeviceObjects();
	}
	
	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::InvalidateDeviceObjects()
{
	// Kill all textures?
	if ( CTextureCache::IsAvailable() )
	{
		CTextureCache::Get()->InvalidateDeviceObjects();
	}
	if ( D3DRender::IsAvailable() )
	{
		D3DRender::Get()->InvalidateDeviceObjects();
	}
	
	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//		 this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT IGraphicsContext::DeleteDeviceObjects()
{
	//RDPCleanup();
	return S_OK;
}
